#!@@PERL@@
#
# Spong client monitoring script.  This runs various tests locally on a machine
# and sends those results to the Spong server machine where the information
# is summarized and displayed.  This script currently checks the following:
#
#    * disk  (check to make sure that disks are not filling up)
#    * cpu   (check to make sure the load is reasonable)
#    * jobs  (check to make sure certain key processes are running)
#    * msgs  (scan through logs looking for problems)
#
# History:
# (1) Ported bb-local.sh script to perl. (Ed Hill Feb 26, 1997)

use lib '@@LIBDIR@@';

use Sys::Hostname;
use Socket;
use POSIX;

use Spong::Daemon;
use Spong::Status qw(status);

srand( time() ^ ($$ + ($$ << 15 )) );

if( $ARGV[0] eq "--debug" )   { $debug = 1;   shift; }
if( $ARGV[0] eq "--restart" ) { $restart = 1; shift; }
if( $ARGV[0] eq "--kill" )    { $kill = 1;    shift; }
if( $ARGV[0] eq "--nosleep" ) { $nosleep = 1; shift; }
if( $ARGV[0] eq "--refresh" ) { $nosleep = 1; shift; }

$me        = "@@BINDIR@@/spong-client";
$conf_file = $ARGV[0] || "@@ETCDIR@@/spong.conf";
($HOST)    = gethostbyname(&Sys::Hostname::hostname());
$HOST      =~ tr/A-Z/a-z/;

$CHECKS = "";

&load_config_files(); # Loads the user specified configuration information
Spong::Daemon::Daemonize  # Daemonize if not signalling or debugging
  unless ($restart || $kill || $nosleep || $debug); 
&handle_signals();    # Set up handlers, and signal the current server if asked

%CHECKFUNCS = (); 
&load_checks();

# Pretty simple program, check the stuff that we are looking at, report it to
# the server, sleep, and do it all again...

while( 1 ) {

   my ($check);
   foreach $check (split / /,$CHECKS) {
      eval { &{$CHECKFUNCS{$check}}(); };
      if ($@) { &error( "Error running check $check: $@") };
   }

   # If we are suppose to stay alive, then sleep about $SPONGSLEEP seconds, we
   # add a little randomness so that things don't get in sync and pound the 
   # spong-server.  Otherwise, just exit.

   if( $nosleep ) {
      last;
   } else {
      my $sleep = int($SPONGSLEEP - (.05 * $SPONGSLEEP) + 
		  rand(.1 * $SPONGSLEEP));
      &debug( "sleeping for $sleep seconds" );
      sleep $sleep;
   }
}

unlink( "$SPONGTMP/spong-client.pid" ) unless $nosleep;
exit(0);



# ===========================================================================
# Utility functions, and signal handlers...
# ===========================================================================

# Load our configuration variables, including anything specific to the host
# that we are running on.

sub load_config_files {
   require $conf_file || die "Can't load $conf_file: $!";
   if( -f "$conf_file.$HOST" ) {
      require "$conf_file.$HOST" || die "Can't load $conf_file.$HOST: $!";
   } else {
      my $tmp = (split( /\./, $HOST ))[0];
      if( -f "$conf_file.$tmp" ) { # for lazy typist
	 require "$conf_file.$tmp" || die "Can't load $conf_file.$tmp: $!";
      }
   }
   &debug( "configuration file(s) loaded" );
}

# This is part of the set up code, this sets up the signal handlers, and
# handles any command line arguments that are given to us that tell us to
# signal the current running spong-server program.

sub handle_signals {

   # Clear out signal mask in case we inherit any blocked sigs

   my $sigset = POSIX::SigSet->new;
   sigprocmask(SIG_SETMASK, $sigset );

   # Set up some signal handlers to handle our death gracefully, and also
   # listen for the HUP signal, and if we se that, we re-exec ourself.

   $SIG{'QUIT'} = \&exit_handler;
   $SIG{'HUP'}  = \&hup_handler;
   $SIG{'USR1'}  = \&hup_handler;
   $SIG{'PIPE'} = \&pipe_handler;

   # If the user gives us the --restart or --kill flags, then we signal the
   # currently running spong-client process, and tell it to either die, or
   # re-exec itself (which causes it to re-read it's configuration files.

   if( $restart || $kill ) {
      open( PID, "$SPONGTMP/spong-client.pid") || die "Can't find pid: $!";
      my $pid = <PID>; chomp $pid;
      close PID;
      
      if( $restart ) { 
	 &debug( "telling pid $pid to restart" ); kill( 'HUP', $pid ); }
      if( $kill ) { 
	 &debug( "telling pid $pid to die" ); kill( 'QUIT', $pid );}
      
      exit(0);
   }

   # Write our pid to the spong tmp directory.
   
   system( "echo $$ >$SPONGTMP/spong-client.pid" ) unless $nosleep;
}


# Output functions, one for debugging information, the other for errors.

sub debug { print scalar localtime, " ", $_[0], "\n" if $main::debug; }
sub error { warn scalar localtime(), " Error: ", $_[0], "\n"; }


# Fork from the parent and become a daemon
sub daemonize {

   my( $pid );

   # Try to fork
   $pid = fork();
   if (! defined $pid) {
      die "ERROR: Could not fork: $!";
   } elsif ($pid) {
      # I'm the parent, so just exit gracefully
      exit(0);
   } else {
      # I'm the child

      # If we are not debuging, disconnect from the console
      if ( ! $debug ) {
         open(STDIN,"</dev/null");
         open(STDOUT,">/dev/null");
         open(STDERR,">/dev/null");
      }

      # Become session group leader
      POSIX::setsid();
   }
}


# Signal handlers...

sub exit_handler { 
   &debug( "caught QUIT signal, exiting..." );
   unlink "$SPONGTMP/spong-client.pid" if -f "$SPONGTMP/spong-client.pid";
   exit(0);
}

sub hup_handler {
   &debug( "caught HUP signal, restarting..." );
   unlink "$SPONGTMP/spong-client.pid" if -f "$SPONGTMP/spong-client.pid";
   $SIG{$_[0]} = \&hup_handler;
   if( $debug ) { exec $me, "--debug"; } else { exec $me; }
   exit(0);
}

sub pipe_handler {
   wait();
   &debug( "caught $_[0] signal." );
   $SIG{$_[0]} = \&pipe_handler;
}

#
# Load client checking functions as defined the $HOSTS checks field
# Default to old checks if 'checks' is not present (cpu disk msgs jobs [local])

sub load_checks {

   if (! defined $CHECKS || ! $CHECKS ) {
      $CHECKS = 'disk cpu processes logs';
      if (defined &check_local) { $CHECKS .= ' local'; }
   };

   foreach $check (split / /,$CHECKS) {
      &debug( "Loading client check $check" );
      eval "require 'Spong/Client/plugins/check_$check';";
      if ( $@ ) { &error( "Could not load $check check plugin" ); }
   }
}
